//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2012 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.IO;
using System.Text;

namespace Smdn.Formats.Mime {
  public class Mail : IDisposable {
    /*
     * class members
     */
    public static Mail Load(string file)
    {
      return new Mail(MimeMessage.Load(file));
    }

    public static Mail Load(string file,
                            Encoding fallbackCharset)
    {
      return new Mail(MimeMessage.Load(file,
                                       fallbackCharset));
    }

    public static Mail Load(string file,
                            EncodingSelectionCallback selectFallbackCharset)
    {
      return new Mail(MimeMessage.Load(file,
                                       selectFallbackCharset));
    }

    public static Mail Load(Stream stream)
    {
      return new Mail(MimeMessage.Load(stream));
    }

    public static Mail Load(Stream stream,
                            Encoding fallbackCharset)
    {
      return new Mail(MimeMessage.Load(stream,
                                       fallbackCharset));
    }

    public static Mail Load(Stream stream,
                            EncodingSelectionCallback selectFallbackCharset)
    {
      return new Mail(MimeMessage.Load(stream,
                                       selectFallbackCharset));
    }

    public static Mail LoadMail(string mail)
    {
      return new Mail(MimeMessage.LoadMessage(mail));
    }

    public static Mail LoadMail(string mail,
                                Encoding fallbackCharset)
    {
      return new Mail(MimeMessage.LoadMessage(mail,
                                              fallbackCharset));
    }

    public static Mail LoadMail(string mail,
                                EncodingSelectionCallback selectFallbackCharset)
    {
      return new Mail(MimeMessage.LoadMessage(mail,
                                              selectFallbackCharset));
    }

    /*
     * instance members
     */
    public MimeMessage BaseMessage {
      get { CheckDisposed(); return baseMessage; }
    }

    public MimeHeaderCollection Headers {
      get { CheckDisposed(); return baseMessage.Headers; }
    }

    public MimeType MimeType {
      get { CheckDisposed(); return baseMessage.MimeType; }
    }

    public MimeMessageDisposition Disposition {
      get { CheckDisposed(); return baseMessage.Disposition; }
    }

    public string Subject {
      get { CheckDisposed(); return GetHeaderValueOf("Subject"); }
    }

    public string From {
      get { CheckDisposed(); return GetHeaderValueOf("From"); }
    }

    public string To {
      get { CheckDisposed(); return GetHeaderValueOf("To"); }
    }

    public string DateString {
      get { CheckDisposed(); return GetHeaderValueOf("Date"); }
    }

    public DateTimeOffset? Date {
      get {
        CheckDisposed();

        try {
          return DateTimeConvert.FromRFC822DateTimeOffsetStringNullable(GetHeaderValueOf("Date"));
        }
        catch (FormatException) {
          return null;
        }
      }
    }

    public string MessageId {
      get { CheckDisposed(); return GetHeaderValueOf("Message-ID"); }
    }

    public string InReplyTo {
      get { CheckDisposed(); return GetHeaderValueOf("In-Reply-To"); }
    }

    public string References {
      get { CheckDisposed(); return GetHeaderValueOf("References"); }
    }

    /// <returns>header value of X-Mailer or User-Agent</returns>
    public string UserAgent {
      get { CheckDisposed(); return GetHeaderValueOf("X-Mailer") ?? GetHeaderValueOf("User-Agent"); }
    }

    public string ReplyTo {
      get { CheckDisposed(); return GetHeaderValueOf("Reply-To"); }
    }

    public string Sender {
      get { CheckDisposed(); return GetHeaderValueOf("Sender"); }
    }

#if false
    public string[] CC {
      get { return GetHeaderValueOf("CC").Split(','); }
    }

    public string[] Bcc {
      get { return GetHeaderValueOf("BCC").Split(','); }
    }
#endif

    public ReadOnlyCollection<MimeMessage> SubParts {
      get { CheckDisposed(); return baseMessage.SubParts; }
    }

    public IEnumerable<MimeMessage> Attachments {
      get {
        CheckDisposed();

        return baseMessage.SubParts.Where(delegate(MimeMessage subPart) {
          return subPart.Disposition == MimeMessageDisposition.Attachment;
        });
      }
    }

    public IEnumerable<MimeMessage> InlineMessages {
      get {
        CheckDisposed();

        return baseMessage.SubParts.Where(delegate(MimeMessage subPart) {
          return subPart.Disposition == MimeMessageDisposition.Inline;
        });
      }
    }

    public Mail(MimeMessage baseMessage)
    {
      if (baseMessage == null)
        throw new ArgumentNullException("baseMessage");

      this.baseMessage = baseMessage;
    }

    public void Dispose()
    {
      if (baseMessage != null) {
        baseMessage.Dispose();
        baseMessage = null;
      }
    }

    private void CheckDisposed()
    {
      if (baseMessage == null)
        throw new ObjectDisposedException(GetType().FullName);
    }

    public string GetHeaderValueOf(string headerName)
    {
      CheckDisposed();

      var header = baseMessage.Headers.GetHeaders(headerName).FirstOrDefault();

      if (header == null)
        return null;
      else
        return header.Value;
    }

    public void ReadContent(Action<StreamReader> read)
    {
      CheckDisposed();

      baseMessage.ReadContent(read);
    }

    public void ReadContent(Action<BinaryReader> read)
    {
      CheckDisposed();

      baseMessage.ReadContent(read);
    }

    public string ReadContentAsText()
    {
      CheckDisposed();

      return baseMessage.ReadContentAsText();
    }

    public byte[] ReadContentAsBinary()
    {
      CheckDisposed();

      return baseMessage.ReadContentAsBinary();
    }

#if false
    public override string ToString()
    {
      CheckDisposed();

      return baseMessage.ToString();
    }
#endif

    private MimeMessage baseMessage;
  }
}

