// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  public class MimeMessage : IDisposable {
    public MimeHeaderCollection Headers {
      get { CheckDisposed(); return headers; }
    }

    internal ChunkedMemoryStream Content {
      get { CheckDisposed(); return content; }
      set { content = value; }
    }

    public ReadOnlyCollection<MimeMessage> SubParts {
      get { CheckDisposed(); return subParts; }
    }

    public MimeType MimeType {
      get; internal set;
    }

    public Encoding Charset {
      get; internal set;
    }

    public string Boundary {
      get; internal set;
    }

    public ContentTransferEncodingMethod TransferEncoding {
      get; internal set;
    }

    public MimeMessageDisposition Disposition {
      get; internal set;
    }

    internal MimeMessage(MimeHeaders headers)
    {
      this.headers = new MimeHeaderCollection(headers);
      this.content = null;
      this.MimeType = MimeType.TextPlain;
      this.Charset = Charsets.ISO8859_1;
      this.Boundary = null;
      this.TransferEncoding = ContentTransferEncodingMethod.SevenBit;
      this.Disposition = MimeMessageDisposition.None;

      SetSubParts(null);
    }

    internal void SetSubParts(IEnumerable<MimeMessage> subParts)
    {
      if (subParts == null)
        this.subParts = (new List<MimeMessage>()).AsReadOnly(); // empty list
      else
        this.subParts = subParts.ToList().AsReadOnly();
    }

    public void Dispose()
    {
      if (subParts != null) {
        foreach (var subPart in subParts) {
          subPart.Dispose();
        }

        subParts = null;
      }

      if (content != null) {
        content.Dispose();
        content = null;
      }

      disposed = true;
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    public static MimeMessage Load(string file)
    {
      return Load(file, (EncodingSelectionCallback)null);
    }

    public static MimeMessage Load(string file,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return Load(file, delegate(string charsetName) {
        return fallbackCharset;
      });
    }

    public static MimeMessage Load(string file,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
      using (var stream = File.OpenRead(file)) {
        return Load(stream, selectFallbackCharset);
      }
    }

    public static MimeMessage LoadMessage(string message)
    {
      return LoadMessage(message, (EncodingSelectionCallback)null);
    }

    public static MimeMessage LoadMessage(string message,
                                          Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return LoadMessage(message, delegate(string charsetName) {
        return fallbackCharset;
      });
    }

    public static MimeMessage LoadMessage(string message,
                                          EncodingSelectionCallback selectFallbackCharset)
    {
      const int lengthThreshold = 20480;

      if (message == null)
        throw new ArgumentNullException("message");

      Stream messageStream = null;

      try {
        if (message.Length <= lengthThreshold) {
          messageStream = new MemoryStream(Charsets.ISO8859_1.GetBytes(message));
        }
        else {
          messageStream = new ChunkedMemoryStream();

          var writer = new StreamWriter(messageStream, Charsets.ISO8859_1);

          writer.Write(message);
          writer.Flush();
          // writer.Close(); // leave stream open

          messageStream.Position = 0L;
        }

        return Load(messageStream,
                    selectFallbackCharset);
      }
      finally {
        if (messageStream != null)
          messageStream.Close();
      }
    }

    public static MimeMessage Load(Stream stream)
    {
      return Load(stream, (EncodingSelectionCallback)null);
    }

    public static MimeMessage Load(Stream stream,
                                   Encoding fallbackCharset)
    {
      if (fallbackCharset == null)
        throw new ArgumentNullException("fallbackCharset");

      return Load(stream, delegate(string charsetName) {
        return fallbackCharset;
      });
    }

    public static MimeMessage Load(Stream stream,
                                   EncodingSelectionCallback selectFallbackCharset)
    {
#if false
      // MIME-Version
      if (message.Headers.Contains("MIME-Version") && message.Headers["MIME-Version"].Value.Trim() != "1.0")
        throw new NotSupportedException("unsupported MIME version");
#endif

      return Parser.Parse(stream, selectFallbackCharset);
    }

    public void ReadContent(Action<StreamReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentTextReader(this));
    }

    public void ReadContent(Action<BinaryReader> read)
    {
      CheckDisposed();

      if (read == null)
        throw new ArgumentNullException("read");

      read(Decoder.CreateContentBinaryReader(this));
    }

    public string ReadContentAsText()
    {
      CheckDisposed();

      return Decoder.CreateContentTextReader(this).ReadToEnd();
    }

    public byte[] ReadContentAsBinary()
    {
      CheckDisposed();

      var contentStream = Decoder.CreateContentReadingStream(this);
      var resultStream = new MemoryStream((int)contentStream.Length);

      contentStream.CopyTo(resultStream);

      resultStream.Close();

      return resultStream.ToArray();
    }

#if false
    public override string ToString()
    {
      CheckDisposed();

      // TODO: impl
      using (var builder = new MimeMessageBuilder(this)) {
        return builder.ToString();
      }
    }
#endif

    private bool disposed = false;
    private /*readonly*/ MimeHeaderCollection headers;
    private ChunkedMemoryStream content;
    private /*readonly*/ ReadOnlyCollection<MimeMessage> subParts;
  }
}