// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2012 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.Text;

using Smdn.IO;

namespace Smdn.Formats.Mime {
 internal static class Parser {
    internal static MimeMessage Parse(Stream stream,
                                      EncodingSelectionCallback selectFallbackCharset)
    {
      return Parse(new LooseLineOrientedStream(stream), selectFallbackCharset);
    }

    private static MimeMessage Parse(LineOrientedStream stream,
                                     EncodingSelectionCallback selectFallbackCharset)
    {
      var headers = MimeUtils.ParseHeader(stream).Select(delegate(KeyValuePair<string, string> pair) {
        MimeEncodingMethod encoding;
        Encoding charset;

        var headerValue = MimeEncoding.Decode(pair.Value,
                                              selectFallbackCharset,
                                              out encoding,
                                              out charset);

        return new MimeHeader(pair.Key, headerValue, encoding, charset);
      });

      return ParseBody(stream,
                       new MimeHeaders(headers),
                       selectFallbackCharset);
    }

    private static MimeMessage ParseBody(LineOrientedStream stream,
                                         MimeHeaders headers,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      var message = new MimeMessage(headers);

      ParseContentType(message, selectFallbackCharset);
      ParseContentTransferEncoding(message);
      ParseContentDisposition(message);

      // read and parse content
      ChunkedMemoryStream contentStream;

      if (message.MimeType == null || !message.MimeType.TypeEquals("multipart")) {
        contentStream = new ChunkedMemoryStream();

        stream.CopyTo(contentStream);

        message.Content = contentStream;

        return message;
      }

      // multipart/*
      var parts = new List<MimeMessage>();
      var delimiter = ByteString.CreateImmutable("--" + message.Boundary);
      var closeDelimiter = ByteString.CreateImmutable("--" + message.Boundary + "--");
      ChunkedMemoryStream body = null;
      ByteString line = null;
      ByteString lastLine = null;

      contentStream = new ChunkedMemoryStream();

      for (;;) {
        if (lastLine != null)
          contentStream.Write(lastLine.Segment.Array,
                              lastLine.Segment.Offset,
                              lastLine.Segment.Count);

        var l = stream.ReadLine();

        if (l == null)
          break;

        lastLine = line;
        line = ByteString.CreateImmutable(l);

        if (line.StartsWith(delimiter)) {
          if (lastLine != null) {
            if (ByteString.IsTerminatedByCRLF(lastLine))
              // CRLF "--" boundary
              contentStream.Write(lastLine.Segment.Array,
                                  lastLine.Segment.Offset,
                                  lastLine.Segment.Count - 2);
            else
              // LF "--" boundary or CR "--" boundary
              contentStream.Write(lastLine.Segment.Array,
                                  lastLine.Segment.Offset,
                                  lastLine.Segment.Count - 1);
          }

          contentStream.Position = 0;

          if (body == null)
            body = contentStream;
          else
            parts.Add(Parse(contentStream, selectFallbackCharset));

          if (line.StartsWith(closeDelimiter))
            break;
          else
            contentStream = new ChunkedMemoryStream();

          lastLine = null;
        }
      }

      message.Content = body;
      message.SetSubParts(parts);

      return message;
    }

    private static void ParseContentType(MimeMessage message,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      if (!message.Headers.Contains(MimeHeaderNames.ContentType))
        return;

      var contentType = message.Headers[MimeHeaderNames.ContentType];
      var mimeTypeString = contentType.GetValueWithoutParameter();

      if (!string.IsNullOrEmpty(mimeTypeString))
        message.MimeType = new MimeType(mimeTypeString);

      var charsetString = contentType.GetParameter("charset", true);

      if (!string.IsNullOrEmpty(charsetString))
        message.Charset = Charsets.FromString(charsetString, selectFallbackCharset);

      message.Boundary = contentType.GetParameter("boundary", true);
    }

    private static void ParseContentTransferEncoding(MimeMessage message)
    {
      message.TransferEncoding = ContentTransferEncodingMethod.SevenBit; // as default

      if (!message.Headers.Contains(MimeHeaderNames.ContentTransferEncoding))
        return;

      var contentTransferEncoding = message.Headers[MimeHeaderNames.ContentTransferEncoding].GetValueWithoutParameter();

      if (string.IsNullOrEmpty(contentTransferEncoding))
        return;

      message.TransferEncoding = ContentTransferEncoding.GetEncodingMethod(contentTransferEncoding);
    }

    private static void ParseContentDisposition(MimeMessage message)
    {
      message.Disposition = MimeMessageDisposition.None; // as default

      if (!message.Headers.Contains(MimeHeaderNames.ContentDisposition))
        return;

      var contentDisposition = message.Headers[MimeHeaderNames.ContentDisposition].GetValueWithoutParameter();

      if (string.IsNullOrEmpty(contentDisposition))
        return;

      switch (contentDisposition.ToLowerInvariant()) {
        case "inline": message.Disposition = MimeMessageDisposition.Inline; break;
        case "attachment": message.Disposition = MimeMessageDisposition.Attachment; break;
        default:
          throw new NotSupportedException(string.Format("unsupported content disposition: '{0}'", contentDisposition));
      }
    }
  }
}