using System;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MailTests {
    [Test]
    public void TestProperties()
    {
      var input = @"MIME-Version: 1.0
Return-Path: <return-path@gmail.com>
Received: from foo by bar with buz
Message-Id: <650A339D.4500208@example.com>
Date: Sun, 16 Sep 2012 16:45:11 +0900
From: from-address <from@example.com>
To: to-address <to@example.com>
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=
In-Reply-To: <CpN_A3fDtNWvrvegSSVUHpQcw@example.com>
References: <CpN_A3fDtNWvrvegSSVUHpQcw@example.com>
User-Agent: Mozilla/5.0 Thunderbird/14.0
Reply-To: reply-to-address <reply-to@example.com>
Sender: sender-address <sender@example.com>
CC: cc1@example.com, cc2@example.com
BCC: bcc1@example.com, bcc2@example.com
Content-Type: multipart/alternative;
  boundary=""==boundary==""
Content-Transfer-Encoding: 7bit

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--";

      var mail = Mail.LoadMail(input);

      Assert.IsNotNull(mail.BaseMessage);
      Assert.AreEqual(17, mail.Headers.Count);
      Assert.AreEqual(MimeType.MultipartAlternative, mail.MimeType);
      Assert.AreEqual(MimeMessageDisposition.None, mail.Disposition);

      Assert.AreEqual("件名", mail.Subject, "Subject");
      Assert.AreEqual("from-address <from@example.com>", mail.From, "From");
      Assert.AreEqual("to-address <to@example.com>", mail.To, "To");
      Assert.AreEqual("Sun, 16 Sep 2012 16:45:11 +0900", mail.DateString, "DateString");
      Assert.AreEqual(new DateTimeOffset(2012, 9, 16, 16, 45, 11, new TimeSpan(+9, 0, 0)), mail.Date, "Date");
      Assert.AreEqual("<650A339D.4500208@example.com>", mail.MessageId, "MessageId");
      Assert.AreEqual("<CpN_A3fDtNWvrvegSSVUHpQcw@example.com>", mail.InReplyTo, "InReplyTo");
      Assert.AreEqual("<CpN_A3fDtNWvrvegSSVUHpQcw@example.com>", mail.References, "References");
      Assert.AreEqual("Mozilla/5.0 Thunderbird/14.0", mail.UserAgent, "UserAgent");
      Assert.AreEqual("reply-to-address <reply-to@example.com>", mail.ReplyTo, "ReplyTo");
      Assert.AreEqual("sender-address <sender@example.com>", mail.Sender, "Sender");
      //Assert.AreEqual("cc1@example.com, cc2@example.com", mail.CC, "CC");
      //Assert.AreEqual("bcc1@example.com, bcc2@example.com", mail.Bcc, "Bcc");

      Assert.AreEqual(2, mail.SubParts.Count, "SubParts.Count");

      CollectionAssert.IsEmpty(mail.Attachments, "Attachments");
      CollectionAssert.IsEmpty(mail.InlineMessages, "InlineMessages");
    }

    [Test]
    public void TestDispose()
    {
      var input = @"MIME-Version: 1.0";

      var mail = Mail.LoadMail(input);

      mail.Dispose();

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.BaseMessage));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Headers));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.SubParts));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.MimeType));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Disposition));

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Subject), "Subject");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.From), "From");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.To), "To");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Date), "Date");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.DateString), "Date");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.MessageId), "MessageId");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.InReplyTo), "InReplyTo");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.References), "References");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.UserAgent), "UserAgent");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.ReplyTo), "ReplyTo");
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Sender), "Sender");
      //Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.CC), "CC");
      //Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mail.Bcc), "Bcc");

      Assert.Throws<ObjectDisposedException>(() => mail.Attachments.ToArray(), "Attachments");
      Assert.Throws<ObjectDisposedException>(() => mail.InlineMessages.ToArray(), "InlineMessages");

      Assert.Throws<ObjectDisposedException>(() => mail.GetHeaderValueOf("Content-Type"));

      Assert.DoesNotThrow(() => mail.Dispose(), "re-Dispose");
    }

    [Test]
    public void TestGetHeaderValueOf()
    {
      var input = @"MIME-Version: 1.0
Subject: foo
Subject: bar";

      var mail = Mail.LoadMail(input);

      Assert.AreEqual("foo", mail.GetHeaderValueOf("SUBJECT"));
      Assert.IsNull(mail.GetHeaderValueOf("X-NonExistent"));
    }
  }
}