using System;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.IO;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeMessageTests {
    [Test/*, ExpectedException(typeof(NotSupportedException))*/]
    public void TestLoadMessageUnsupportedMimeVersion()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1
Content-Type: text/plain; charset=""iso-2022-jp""
Content-Transfer-Encoding: quoted-printable
");
    }

    [Test]
    public void TestLoadMessageMimeVersionNotExist()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
    }

    [Test/*, ExpectedException(typeof(NotSupportedException))*/]
    public void TestLoadMessageMimeVersionNotSupported()
    {
      MimeMessage.LoadMessage(@"MIME-Version: 1.1\r\n");
    }

    [Test]
    // TODO: [ExpectedException(typeof(NotSupportedException))]
    public void TestLoadMessageMimeVersionInvalid()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0 (comment)\"\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionWithComment()
    {
      MimeMessage.LoadMessage("MIME-Version: 1.0 (comment)\r\n");
    }

    [Test]
    public void TestLoadMessageMimeVersionQuoted()
    {
      MimeMessage.LoadMessage("MIME-Version: \"1.0\"\r\n");
    }

    [Test]
    public void TestLoadMessageFromStringLargeMessage()
    {
      var header = "Content-Type: text/plain\r\n\r\n";
      var body = new StringBuilder();

      var line = "0123456789"; // 10

      line += line; // 20
      line += line; // 40
      line += line; // 80

      while (body.Length <= 40960) {
        body.AppendLine(line);
      }

      var mime = MimeMessage.LoadMessage(header + body);

      Assert.AreEqual(body.ToString(),
                      mime.ReadContentAsText());
    }

    [Test]
    public void TestLoadMessageSinglePartTextBody()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain
Subject: single part

single-part text body");

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("single part", mime.Headers["Subject"].Value);
      Assert.AreEqual("single-part text body", mime.ReadContentAsText());
    }

    [Test]
    public void TestLoadMessageMultiPartTextBody()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--");

      Assert.AreEqual(MimeType.MultipartAlternative, mime.MimeType);
      Assert.AreEqual("==boundary==", mime.Boundary);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("multipart/alternative; boundary=\"==boundary==\"",
                      mime.Headers["Content-Type"].Value);
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("multi part", mime.Headers["Subject"].Value);
      Assert.AreEqual("multipart message body\n", mime.ReadContentAsText().Replace("\r\n", "\n"));

      Assert.AreEqual(2, mime.SubParts.Count);

      Assert.AreEqual(MimeType.TextPlain, mime.SubParts[0].MimeType);
      Assert.IsTrue(mime.SubParts[0].Headers.Contains("Content-Type"));
      Assert.AreEqual("text/plain", mime.SubParts[0].Headers["Content-Type"].Value);
      Assert.AreEqual("first part, text/plain\n", mime.SubParts[0].ReadContentAsText().Replace("\r\n", "\n"));

      Assert.AreEqual(new MimeType("text", "html"), mime.SubParts[1].MimeType);
      Assert.IsTrue(mime.SubParts[1].Headers.Contains("Content-Type"));
      Assert.AreEqual("text/html", mime.SubParts[1].Headers["Content-Type"].Value);
      Assert.AreEqual("<html><body><p>second part, text/html</p></body></html>",
                      mime.SubParts[1].ReadContentAsText());
    }
    
    [Test]
    public void TestLoadMessageMultiPartWithNoHeader()
    {
      var mime = MimeMessage.LoadMessage(@"Return-Path: <>
Date: Fri, 21 Dec 2007 07:04:35 +0900
MIME-Version: 1.0
Content-Type: multipart/report; report-type=delivery-status;
  boundary=""lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx""
Subject: Warning: could not send message for past 4 hours
Auto-Submitted: auto-generated (warning-timeout)

This is a MIME-encapsulated message

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx

    **********************************************
    **      THIS IS A WARNING MESSAGE ONLY      **
    **  YOU DO NOT NEED TO RESEND YOUR MESSAGE  **
    **********************************************

The original message was received at Fri, 21 Dec 2007 02:06:24 +0900
from xxx.xxxxxx.xxx [127.0.0.1]

   ----- Transcript of session follows -----
451 xxx.xxxxxx.xxx: Name server timeout
Warning: message still undelivered after 4 hours
Will keep trying until message is 5 days old
451 xxx.xxxxxx.xxx: Name server timeout

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/delivery-status

Reporting-MTA: dns; xxx.xxxxxx.xxx
Arrival-Date: Fri, 21 Dec 2007 02:06:24 +0900

Final-Recipient: RFC822; root@xxx.xxxxxx.xxx
X-Actual-Recipient: RFC822; xxxxxx@xxx.xxxxxx.xxx
Action: delayed
Status: 4.4.3
Last-Attempt-Date: Fri, 21 Dec 2007 07:04:35 +0900
Will-Retry-Until: Wed, 26 Dec 2007 02:06:24 +0900

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx
Content-Type: message/rfc822

Return-Path: <root@xxx.xxxxxx.xxx>
Date: Fri, 21 Dec 2007 02:02:15 +0900
Content-Type: text/plain; charset=UTF-8
Auto-Submitted: auto-generated
X-Cron-Env: <MAILTO=root,postmaster,webmaster,clamav>
X-Cron-Env: <SHELL=/bin/sh>
X-Cron-Env: <HOME=/root>
X-Cron-Env: <PATH=/usr/bin:/bin>
X-Cron-Env: <LOGNAME=root>
X-Cron-Env: <USER=root>

ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Can't query current.cvd.clamav.net
ERROR: Can't get information about database.clamav.net: Temporary DNS error
ERROR: No servers could be reached. Giving up
ERROR: Update failed. Your network may be down or none of the mirrors listed in freshclam.conf is working.

--lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx--
");

      Assert.AreEqual(new MimeType("multipart", "report"), mime.MimeType);
      Assert.AreEqual("lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx", mime.Boundary);
      Assert.IsTrue(mime.Headers.Contains("Content-Type"));
      Assert.AreEqual("multipart/report; report-type=delivery-status; boundary=\"lBKM1GFo005500.1198188275/xxx.xxxxxx.xxx\"",
                      mime.Headers["Content-Type"].Value);

      Assert.AreEqual(3, mime.SubParts.Count);

      Assert.AreEqual(0, mime.SubParts[0].Headers.Count);
      Assert.IsTrue(mime.SubParts[0].ReadContentAsText().StartsWith("    ***"));

      Assert.AreEqual(new MimeType("message", "delivery-status"), mime.SubParts[1].MimeType);
      Assert.AreEqual("message/delivery-status", mime.SubParts[1].Headers["Content-Type"].Value);
      Assert.AreEqual(new MimeType("message", "rfc822"), mime.SubParts[2].MimeType);
      Assert.AreEqual("message/rfc822", mime.SubParts[2].Headers["Content-Type"].Value);
    }

    [Test]
    public void TestHeaders()
    {
      var message =
        "content-type: text/plain\r\n" +
          "FROM: from@example.com\r\n" + 
          "tO: to@example.com\r\n";

      var mime = MimeMessage.LoadMessage(message);

      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.IsTrue(mime.Headers.Contains("CONTENT-TYPE"));
      Assert.AreEqual("text/plain", mime.Headers["Content-Type"].Value);
      Assert.AreSame(mime.Headers["CONTENT-TYPE"], mime.Headers["Content-Type"]);

      Assert.IsTrue(mime.Headers.Contains("from"));
      Assert.AreEqual("from@example.com", mime.Headers["FROM"].Value);

      Assert.IsTrue(mime.Headers.Contains("to"));
      Assert.AreEqual("to@example.com", mime.Headers["TO"].Value);
    }

    [Test]
    public void TestHeadersMultipleHeader()
    {
      var message = @"MIME-Version: 1.0\r\n
Received: #1
Received: #2
Received: #3";

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsTrue(mime.Headers.Contains("Received"));

      Assert.AreEqual(3, mime.Headers.GetHeaders("Received").Count());
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#1").Count(), "#1");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#2").Count(), "#2");
      Assert.AreEqual(1, mime.Headers.GetHeaders("Received").Where(h => h.Value == "#3").Count(), "#2");
    }

    [Test]
    public void TestDispose()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain

first part, text/plain

--==boundary==
Content-Type: text/html

<html><body><p>second part, text/html</p></body></html>
--==boundary==--");

      var part1 = mime.SubParts[0];
      var part2 = mime.SubParts[1];

      mime.Dispose();

      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.Headers));
      Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(mime.SubParts));
      Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(StreamReader r) {
        r.ReadLine();
      }));
      Assert.Throws<ObjectDisposedException>(() => mime.ReadContent(delegate(BinaryReader r) {
        r.ReadByte();
      }));
      Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsText());
      Assert.Throws<ObjectDisposedException>(() => mime.ReadContentAsBinary());

      Assert.Throws<ObjectDisposedException>(() => part1.ReadContentAsText());
      Assert.Throws<ObjectDisposedException>(() => part2.ReadContentAsText());

      Assert.DoesNotThrow(() => mime.Dispose(), "re-Dispose");
    }

    [Test]
    public void TestReadContentAsText()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");

      Assert.AreEqual("content", mime.ReadContentAsText());
      Assert.AreEqual("content", mime.ReadContentAsText(), "read again");
    }

    [Test]
    public void TestReadContentAsBinary()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");

      Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary());
      Assert.AreEqual(Charsets.Latin1.GetBytes("content"), mime.ReadContentAsBinary(), "read again");
    }

    [Test]
    public void TestReadContentContentTypeNotExist()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Subject: transfer encoding not exist

decode as text/plain; charset=us-ascii");
      
      Assert.AreEqual("decode as text/plain; charset=us-ascii", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestReadContentContentTypeNonText()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: application/octet-stream
Subject: transfer encoding not exist

decode as binary stream");
      
      var text = Encoding.ASCII.GetBytes("decode as binary stream");
      
      Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
      Assert.AreEqual(text, mime.ReadContentAsBinary());
    }
    
    [Test]
    public void TestReadContentContentTransferEncoding7BitMessage()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: 7bit transfer encoding

7bit transfer encoding");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ASCII, mime.Charset);
      Assert.AreEqual(ContentTransferEncodingMethod.SevenBit, mime.TransferEncoding);
      Assert.AreEqual("7bit", mime.Headers["Content-Transfer-Encoding"].Value);
      Assert.AreEqual("7bit transfer encoding", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestReadContentContentTransferEncoding8BitMessage()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=iso-8859-1
Content-Transfer-Encoding: 8bit
Subject: 8bit transfer encoding

" + "8bit transfer encoding \u0080 \u00cf \u00ff\r\n");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.Latin1, mime.Charset);
      Assert.AreEqual(ContentTransferEncodingMethod.EightBit, mime.TransferEncoding);
      Assert.AreEqual("8bit", mime.Headers["Content-Transfer-Encoding"].Value);
      Assert.AreEqual("8bit transfer encoding \u0080 \u00cf \u00ff\r\n", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestReadContentContentTransferEncodingBinaryMessage()
    {
      var message = new MemoryStream(0x200);
      
      var header = Charsets.Latin1.GetBytes(@"MIME-Version: 1.0
Content-Type: application/octet-stream
Content-Transfer-Encoding: binary
Subject: binary transfer encoding

");
      
      message.Write(header, 0, header.Length);
      
      var bodyStream = new MemoryStream(); 
      
      for (var i = 0; i < 0x100; i++) {
        bodyStream.WriteByte((byte)i);
      }
      
      var body = bodyStream.ToArray();
      
      message.Write(body, 0, body.Length);
      
      var mime = MimeMessage.Load(new MemoryStream(message.ToArray()));
      
      Assert.AreEqual(MimeType.ApplicationOctetStream, mime.MimeType);
      Assert.AreEqual(ContentTransferEncodingMethod.Binary, mime.TransferEncoding);
      Assert.AreEqual("binary", mime.Headers["Content-Transfer-Encoding"].Value);
      
      Assert.AreEqual(body, mime.ReadContentAsBinary());
    }
    
    [Test]
    public void TestReadContentContentTransferEncodingBase64Message()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=shift_jis
Content-Transfer-Encoding: base64
Subject: base64 transfer encoding

ir+OmmFiY4KpgsgxMjODSoNpDQqCUIJRglKCgYKCgoMNCg==");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
      Assert.AreEqual(ContentTransferEncodingMethod.Base64, mime.TransferEncoding);
      Assert.AreEqual("base64", mime.Headers["Content-Transfer-Encoding"].Value);
      Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestReadContentContentTransferEncodingQuotedPrintableMessage()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: quoted-printable
Subject: quoted printable transfer encoding

=B4=C1=BB=FAabc=A4=AB=A4=CA123=A5=AB=A5=CA=0D=0A=A3=B1=A3=B2=A3=B3=A3=E1=A3=
=E2=A3=E3=0D=0A");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
      Assert.AreEqual(ContentTransferEncodingMethod.QuotedPrintable, mime.TransferEncoding);
      Assert.AreEqual("quoted-printable", mime.Headers["Content-Transfer-Encoding"].Value);
      Assert.AreEqual("漢字abcかな123カナ\r\n１２３ａｂｃ\r\n", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestReadContentContentTransferEncodingUnsupported()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=euc-jp
Content-Transfer-Encoding: unsupported
Subject: unsupported transfer encoding

unsupported transfer encoding");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
      Assert.AreEqual(ContentTransferEncodingMethod.Unknown, mime.TransferEncoding);
      Assert.AreEqual("unsupported", mime.Headers["Content-Transfer-Encoding"].Value);
      
      try {
        mime.ReadContentAsText();
        
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }
    }

    [Test]
    public void TestReadContentStreamReader()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");

      mime.ReadContent(delegate(StreamReader reader) {
        Assert.AreEqual(Charsets.Latin1, reader.CurrentEncoding);
        Assert.AreEqual("content", reader.ReadToEnd());
      });

      mime.ReadContent(delegate(StreamReader reader) {
        Assert.AreEqual("content", reader.ReadToEnd(), "read again");
      });
    }

    [Test]
    public void TestReadContentBinaryReader()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");

      mime.ReadContent(delegate(BinaryReader reader) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                        reader.ReadBytes(7));
        Assert.AreEqual(-1, reader.PeekChar());
      });

      mime.ReadContent(delegate(BinaryReader reader) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                        reader.ReadBytes(7),
                        "read again");
        Assert.AreEqual(-1, reader.PeekChar());
      });
    }

    [Test]
    public void TestReadContentStreamReaderCloseReaderPlainEncoding()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");
      
      mime.ReadContent(delegate(StreamReader reader) {
        reader.Close();
      });
      
      mime.ReadContent(delegate(StreamReader reader) {
        Assert.AreEqual("content", reader.ReadToEnd(), "read again");
      });
    }

    [Test]
    public void TestReadContentStreamReaderCloseReaderNonPlainEncoding()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==");
      
      mime.ReadContent(delegate(StreamReader reader) {
        reader.Close();
      });
      
      mime.ReadContent(delegate(StreamReader reader) {
        Assert.AreEqual("content", reader.ReadToEnd(), "read again");
      });
    }
    
    [Test]
    public void TestReadContentBinaryReaderCloseReaderPlainEncoding()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain
Subject: test

content");
      
      mime.ReadContent(delegate(BinaryReader reader) {
        reader.Close();
      });
      
      mime.ReadContent(delegate(BinaryReader reader) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                        reader.ReadBytes(7),
                        "read again");
        Assert.AreEqual(-1, reader.PeekChar());
      });
    }

    [Test]
    public void TestReadContentBinaryReaderCloseReaderNonPlainEncoding()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Type: text/plain
Subject: test

Y29udGVudA==");
      
      mime.ReadContent(delegate(BinaryReader reader) {
        reader.Close();
      });

      mime.ReadContent(delegate(BinaryReader reader) {
        Assert.AreEqual(Charsets.Latin1.GetBytes("content"),
                        reader.ReadBytes(7),
                        "read again");
        Assert.AreEqual(-1, reader.PeekChar());
      });
    }
  }
}