using System;
#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using System.IO;
using System.Text;
using NUnit.Framework;

namespace Smdn.Formats.Mime {
  [TestFixture]
  public class MimeMessageParsingTests {
    [Test]
    public void TestParseHeaderContentTypeCharsetContainsLeadingWhitespaces()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset= iso-2022-jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetContainsTrailingWhitespaces()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=iso-2022-jp 
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuoted()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=""iso-2022-jp""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuotedContainsWhitespaces1()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset="" iso-2022-jp ""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetQuotedContainsWhitespaces2()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=""shift jis""
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal1()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=shift-jis
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.ShiftJIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal2()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=euc_jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Encoding.GetEncoding("euc-jp"), mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }
    
    [Test]
    public void TestParseHeaderContentTypeCharsetIllegal3()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=iso_2022-jp
Subject: test

content");
      
      Assert.AreEqual(MimeType.TextPlain, mime.MimeType);
      Assert.AreEqual(Charsets.JIS, mime.Charset);
      Assert.AreEqual("content", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetNotFoundOrUnsupported()
    {
      try {
        MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body");

        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException) {
      }
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetUseFallbackCharset()
    {
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", Encoding.UTF8);

      Assert.IsNotNull(mime.Charset);
      Assert.AreEqual(Encoding.UTF8, mime.Charset);
      Assert.AreEqual("body", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharset()
    {
      var called = false;
      var mime = MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", delegate(string name) {
        called = true;

        Assert.AreEqual("x-unsupported", name, "callback name arg");

        return Encoding.UTF8;
      });

      Assert.IsTrue(called);
      Assert.IsNotNull(mime.Charset);
      Assert.AreEqual(Encoding.UTF8, mime.Charset);
      Assert.AreEqual("body", mime.ReadContentAsText());
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharsetReturnNull()
    {
      var called = false;

      try {
        MimeMessage.LoadMessage(@"Content-Type: text/plain; charset=x-unsupported

body", delegate(string name) {
          called = true;
          return null;
        }
        );

        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException ex) {
        Assert.AreEqual("x-unsupported", ex.EncodingName, "EncodingName");
      }

      Assert.IsTrue(called);
    }

    [Test]
    public void TestParseHeaderContentTypeCharsetSelectFallbackCharsetWithLoadFromFile()
    {
      const string testFile = "test.eml";

      try {
        if (File.Exists(testFile))
          File.Delete(testFile);

        File.WriteAllText(testFile, @"Content-Type: text/plain; charset=x-unsupported

body", Encoding.ASCII);

        var mime = MimeMessage.Load(testFile, delegate(string name) {
          Assert.AreEqual("x-unsupported", name, "callback name arg");

          return Encoding.UTF8;
        });

        Assert.IsNotNull(mime.Charset);
        Assert.AreEqual(Encoding.UTF8, mime.Charset);
        Assert.AreEqual("body", mime.ReadContentAsText());
      }
      finally {
        if (File.Exists(testFile))
          File.Delete(testFile);
      }
    }

    [Test]
    public void TestParseBodyNewlineDelimitedByInternalStreamBuffer()
    {
      const int bufferSize = 1024;

      var message = "MIME-Version: 1.0\r\nSubject: ";
      var splittedAt = bufferSize - message.Length - 1;
      var expectedSubject = new string('x', splittedAt) + " yyy";

      message += new string('x', splittedAt);
      message += "\r";

      if (message.Length != bufferSize)
        Assert.Fail("not splitted");

      message += "\n";
      message += "\tyyy\r\n";
      message += "From: from@example.com\r\n\r\n";
      message += "body";

      var mime = MimeMessage.LoadMessage(message);

      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual(expectedSubject, mime.Headers["Subject"].Value);

      Assert.IsTrue(mime.Headers.Contains("From"));
      Assert.AreEqual("from@example.com", mime.Headers["From"].Value);

      Assert.AreEqual("body", mime.ReadContentAsText());
    }



    [Test]
    public void TestDecodeHeaderPlainText()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: subject

body");
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("subject", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.None, mime.Headers["Subject"].Encoding);
      Assert.IsNull(mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeader()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=

body");
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderSubPart()
    {
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: multipart/alternative;
  boundary=""==boundary==""
Subject: multi part

multipart message body

--==boundary==
Content-Type: text/plain
Subject: =?ISO-2022-JP?B?GyRCN29MPhsoQg==?=

first part

--==boundary==--");
      
      Assert.IsTrue(mime.SubParts[0].Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.SubParts[0].Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.SubParts[0].Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.SubParts[0].Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderInvalidCharsetSelectFallback()
    {
      bool called = false;
      var mime = MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?invalid?B?GyRCN29MPhsoQg==?=

body", delegate(string name) {
        called = true;
        return Charsets.JIS;
      });
      
      Assert.IsTrue(called);
      
      Assert.IsTrue(mime.Headers.Contains("Subject"));
      Assert.AreEqual("件名", mime.Headers["Subject"].Value);
      Assert.AreEqual(MimeEncodingMethod.BEncoding, mime.Headers["Subject"].Encoding);
      Assert.AreEqual(Charsets.JIS, mime.Headers["Subject"].Charset);
    }
    
    [Test]
    public void TestDecodeHeaderInvalidCharsetSelectFallbackReturnNull()
    {
      bool called = false;
      
      try {
        MimeMessage.LoadMessage(@"MIME-Version: 1.0
Content-Type: text/plain; charset=us-ascii
Content-Transfer-Encoding: 7bit
Subject: =?invalid?B?GyRCN29MPhsoQg==?=

body", delegate(string name) {
          called = true;
          return null;
        }
        );
        
        Assert.Fail("EncodingNotSupportedException not thrown");
      }
      catch (EncodingNotSupportedException ex) {
        Assert.AreEqual("invalid", ex.EncodingName, "EncodingName");
      }
      
      Assert.IsTrue(called);
    }
  }
}